#include <stdlib.h>
#include "ruby.h"
#include "st.h"

void *
ruby_xmalloc(size)
    long size;
{
    if (size == 0) size = 1;
    return malloc(size);
}

void *
ruby_xcalloc(n, size)
    long n, size;
{
    void *mem;

    mem = ruby_xmalloc(n * size);
    memset(mem, 0, n * size);

    return mem;
}

void *
ruby_xrealloc(ptr, size)
    void *ptr;
    long size;
{
    if (!ptr) return xmalloc(size);
    return realloc(ptr, size);
}

typedef long code_t;

struct method_body {
    VALUE *pool;
    code_t *code;
    VALUE(*func)();
};

struct env_body {
    VALUE *pool;
    code_t *return_addr;
    struct env_body *prev;
    long *sp;
    long *locals;
};
struct env_body *env_cache;

st_table *method_table;

#define CACHE_SIZE 0x80
static struct cache_entry {
    int id;
    struct method_body *body;
} cache[CACHE_SIZE];

inline struct method_body*
lookup_method(VALUE self, int n)
{
    struct cache_entry *ent;
    struct method_body *body;

    ent = cache + n % CACHE_SIZE;
    if (ent && ent->id == n) return ent->body;
    st_lookup(method_table, n, &body);
    cache[n % CACHE_SIZE].id = n;
    cache[n % CACHE_SIZE].body = body;
    return body;
}

/**
STOP		󥿥ץ꥿ߤ롣
DEFUN i	n	poolnܤΥǡ᥽åi롣
FCALL n		᥽åɤƤӽФnϥ᥽å̾(ID)
RETURN		᥽åɤλ롣
JMP_FALSE n	accʤ饸פ롣nդ
POP		åƬacc롣
PUSH_ACC	acc쥸ͤ򥹥å˥ץå夹롣
PUSH_SELF	self쥸ͤ򥹥å˥ץå夹롣
PUSH_LIT n	poolnܤΥƥͤ򥹥å˥ץå夹롣
PUSH_CONST n	n򥹥å˥ץå夹롣
PUSH_LVAR n	nܤΥѿͤ򥹥å˥ץå夹롣
LOAD_LVAR n	nܤΥѿͤacc롣
LOAD_INT n	nacc롣
**/

enum codetype {
    STOP = 0,
    RETURN,
    DEFUN,
    FCALL,
    JMP_FALSE,
    POP,
    PUSH_ACC,
    PUSH_SELF,
    PUSH_LIT,
    PUSH_CONST,
    PUSH_LVAR,
    LOAD_LVAR,
    LOAD_INT,
};

#if defined(__GNUC__) && __GNUC__ >= 2 && !defined(DEBUG)
#define DIRECT_JUMP
#endif

#ifdef DIRECT_JUMP
# define Instr(name) label_##name
# define Next goto *jumptable[*pc++]
#else
# define Instr(name) case name
# define Next break
#endif

#define STACK_SIZE 102400

int stack_size = STACK_SIZE;

#define PUSH(n) do {\
    if (sp >= sp_end) {\
	long offset = sp - sp_beg;\
	stack_size += STACK_SIZE;\
	sp_beg = ruby_xrealloc(sp_beg, stack_size*sizeof(long));\
	sp = sp_beg + offset;\
    }\
    *sp++ = (n);\
} while (0)

long *sp_beg, *sp_end;

VALUE
interp(VALUE self, struct method_body *top_body)
{
#if defined(__GNUC__) && defined(i386)
    register code_t *pc asm("%esi");
    register long *sp asm("%edi");
#else
    register code_t *pc;
    register long *sp;
#endif
    VALUE *pool = top_body->pool;
    register VALUE *locals;
    register VALUE acc = Qnil;
    struct env_body *env;
    int n;

#ifdef DIRECT_JUMP
    static void *jumptable[] = {
	&&Instr(STOP),
	&&Instr(RETURN),
	&&Instr(DEFUN),
	&&Instr(FCALL),
	&&Instr(JMP_FALSE),
	&&Instr(POP),
	&&Instr(PUSH_ACC),
	&&Instr(PUSH_SELF),
	&&Instr(PUSH_LIT),
	&&Instr(PUSH_CONST),
	&&Instr(PUSH_LVAR),
	&&Instr(LOAD_LVAR),
	&&Instr(LOAD_INT),
    };
#endif

    pc = top_body->code;

    sp_beg = ruby_xmalloc(sizeof(long)*STACK_SIZE);
    sp_end = sp_beg + STACK_SIZE;
    sp = sp_beg;

    env = ruby_xmalloc(sizeof(struct env_body));
    env->return_addr = 0;
    env->prev = 0;
    env->pool = top_body->pool;
    env->sp = env->locals = sp;
    locals = env->locals;

#ifdef DIRECT_JUMP
    Next;
#else
    for (;;) {
	switch (*pc++) {
#endif
	  Instr(STOP): {
	      printf("stop %d\n", FIX2INT(acc));
	      return acc;
	  }

	  Instr(RETURN): {
	      code_t *addr = env->return_addr;
	      struct env_body *tmp = env;

	      pc = addr;
	      sp = env->sp;
	      env = env->prev;
	      tmp->prev = env_cache;
	      env_cache = tmp;
	      pool = env->pool;
	      locals = env->locals;
	      Next;
	  }

	  Instr(DEFUN): {
	      n = *pc++;
	      st_insert(method_table, n, pool[*pc++]);
	      Next;
	  }

	  Instr(FCALL): {
	      struct method_body *m;
	      long *p;

	      n = *pc++;
#if 0
	      switch (n) {
		case '>':
		  sp -= acc;
		  acc = sp[0] >= sp[1] ? Qtrue : Qfalse;
		  Next;
		case '-':
		  sp -= acc;
		  acc = INT2FIX(FIX2INT(sp[0]) - FIX2INT(sp[1]));
		  Next;
		default:
		  p = sp - acc;
		  m = lookup_method(self, n);
#else
	      m = lookup_method(self, n);
	      if (m->func) {
		  sp -= acc;
		  acc = m->func(sp[0], sp[1]);
	      }
	      else {
		  p = sp - acc;
#endif
		  if (*pc == RETURN) { /* tail recursion */
		      *pc++;
		  }
		  else {
		      struct env_body *e;

		      if (env_cache) {
			  e = env_cache;
			  env_cache = e->prev;
		      }
		      else {
			  e = ruby_xmalloc(sizeof(struct env_body));
		      }
		      e->prev = env;
		      env = e;
		      env->return_addr = pc;
		      env->sp = p;
		  }
		  locals = env->locals = p;
		  pool = env->pool = m->pool;
		  pc = m->code;
	      }
	      Next;
	  }

	  Instr(JMP_FALSE): {
	      n = *pc++;
	      if (acc) {
		  Next;
	      }
	      pc += n;
	      Next;
	  }

	  Instr(POP): {
	      acc = *--sp;
	      Next;
	  }

	  Instr(PUSH_ACC): {
	      PUSH(acc);
	      Next;
	  }

	  Instr(PUSH_SELF): {
	      PUSH(self);
	      Next;
	  }
	    
	  Instr(PUSH_LIT): {
	      PUSH(pool[*pc++]);
	      Next;
	  }

	  Instr(PUSH_CONST): {
	      n = *pc++;
	      PUSH(n);
	      Next;
	  }

	  Instr(PUSH_LVAR): {
	      PUSH(locals[*pc++]);
	      Next;
	  }

	  Instr(LOAD_LVAR): {
	      acc = locals[*pc++];
	      Next;
	  }

	  Instr(LOAD_INT): {
	      acc = *pc++;
	      Next;
	  }
#ifndef DIRECT_JUMP
	}
    }
#endif
}

VALUE
fix_minus(VALUE x, VALUE y)
{
    return INT2FIX(FIX2INT(x) - FIX2INT(y));
}

VALUE
fix_ge(VALUE x, VALUE y)
{
    if (FIX2INT(x) >= FIX2INT(y)) {
	return Qtrue;
    }
    return Qfalse;
}

struct method_body minus_body[] = {{
    0, 0, fix_minus
}};

struct method_body ge_body[] = {{
    0, 0, fix_ge
}};

code_t tak_code[] = 
    {PUSH_LVAR, 1, PUSH_LVAR, 0, LOAD_INT, 2, FCALL, '>',
     JMP_FALSE, 3,
       LOAD_LVAR, 2,
       RETURN,
	 PUSH_LVAR, 0, PUSH_CONST, INT2FIX(1), LOAD_INT, 2, FCALL, '-', PUSH_ACC,
	 PUSH_LVAR, 1,
	 PUSH_LVAR, 2,
       LOAD_INT, 3, FCALL, 't',
       PUSH_ACC,
	 PUSH_LVAR, 1, PUSH_CONST, INT2FIX(1), LOAD_INT, 2, FCALL, '-', PUSH_ACC,
	 PUSH_LVAR, 2,
	 PUSH_LVAR, 0,
       LOAD_INT, 3, FCALL, 't',
       PUSH_ACC,
	 PUSH_LVAR, 2, PUSH_CONST, INT2FIX(1), LOAD_INT, 2, FCALL, '-', PUSH_ACC,
	 PUSH_LVAR, 0,
	 PUSH_LVAR, 1,
       LOAD_INT, 3,
       FCALL, 't',
       PUSH_ACC,
     LOAD_INT, 3, FCALL, 't',
     RETURN};
struct method_body tak_body[] = {{
    0, tak_code, 0
}};

code_t test_code[] = 
    {PUSH_CONST, INT2FIX(18),
     PUSH_CONST, INT2FIX(12),
     PUSH_CONST, INT2FIX(6),
     LOAD_INT, 3, FCALL, 't',
     RETURN};
struct method_body test_body[] = {{
    0, test_code, 0
}};

VALUE main_pool[] = {(VALUE)tak_body, (VALUE)test_body};
code_t main_code[] = 
    {DEFUN, 't', 0, DEFUN, 'T', 1,
     LOAD_INT, 0, FCALL, 'T',
     LOAD_INT, 0, FCALL, 'T',
     LOAD_INT, 0, FCALL, 'T',
     LOAD_INT, 0, FCALL, 'T',
     LOAD_INT, 0, FCALL, 'T',
     LOAD_INT, 0, FCALL, 'T',
     LOAD_INT, 0, FCALL, 'T',
     LOAD_INT, 0, FCALL, 'T',
     LOAD_INT, 0, FCALL, 'T',
     LOAD_INT, 0, FCALL, 'T',
     LOAD_INT, 0, FCALL, 'T',
     LOAD_INT, 0, FCALL, 'T',
     LOAD_INT, 0, FCALL, 'T',
     LOAD_INT, 0, FCALL, 'T',
     LOAD_INT, 0, FCALL, 'T',
     LOAD_INT, 0, FCALL, 'T',
     LOAD_INT, 0, FCALL, 'T',
     LOAD_INT, 0, FCALL, 'T',
     LOAD_INT, 0, FCALL, 'T',
     LOAD_INT, 0, FCALL, 'T',
     STOP};
struct method_body main_body[] = {{
    main_pool, main_code, 0
}};

int
main()
{
    method_table = st_init_numtable();
    st_insert(method_table, '-', minus_body);
    st_insert(method_table, '>', ge_body);
    interp(Qnil, main_body);
    exit(0);
}
